
from .__version import __version__

from .variable_type import SPIN, BINARY, Vartype, cast_var_type


from .sampler import Response
from .sampler import SASampler, SQASampler, CSQASampler
from .sampler import GPUChimeraSQASampler, GPUChimeraSASampler, CMOSAnnealer
from .model import BinaryQuadraticModel, KingGraph, ChimeraModel
from .utils import solver_benchmark, convert_response



import time
import numpy as np
from openjij.model import BinaryHigherOrderModel
from openjij import SPIN, BINARY, cast_var_type
import openjij


def default_schedule(bhom, beta_min, beta_max, num_sweeps):

    _beta_max, _beta_min = _default_beta_schedule(bhom)

    beta_max = beta_max if beta_max else _beta_max
    beta_min = beta_min if beta_min else _beta_min

    num_sweeps_per_beta = max(1, num_sweeps // 1000)
    r = (beta_max/beta_min)**(1/num_sweeps)

    return [[beta_min * r ** n, num_sweeps_per_beta] for n in range(num_sweeps)]


def _default_beta_schedule(hubo):
    adj_dict = hubo.adj_dict()
    linear = hubo.interactions[0]

    de_list = []
    for i, adj in adj_dict.items():
        de = linear[i]
        for _, v in adj:
            de += np.abs(v)
        de_list.append(de)

    n = len(linear)
    beta_max = np.log(n)/np.max(de_list)

    abs_linear = np.abs(list(linear.values())) + 0.01
    beta_min = np.log(5)/np.min(abs_linear)

    return beta_min, beta_max


def hubo_simulated_annealing(bhom: BinaryHigherOrderModel, state: list,
                             schedule: list, var_type):

    adj_dict = bhom.adj_dict()
    state = np.array(state)

    if SPIN == cast_var_type(var_type):
        for beta, mc_steps in schedule:
            for _ in range(mc_steps):
                for i in bhom.indices:
                    de = -2*state[i]*np.sum([d*np.prod(state[_inds])
                                             for _inds, d in adj_dict[i]])
                    de += -2*state[i]*bhom.interactions[0][i]
                    if de < 0 or np.random.uniform(0, 1) < np.exp(-beta*de):
                        state[i] *= -1
        return state
    elif BINARY == cast_var_type(var_type):
        for beta, mc_steps in schedule:
            for _ in range(mc_steps):
                for i in bhom.indices:
                    de = (1-2*state[i])*np.sum([d*np.prod(state[_inds])
                                                for _inds, d in adj_dict[i]])
                    de += (1-2*state[i])*bhom.interactions[0][i]
                    if de <= 0 or np.random.uniform(0, 1) < np.exp(-beta*de):
                        state[i] = 0 if state[i] == 1 else 1
        return state
    else:
        raise ValueError("var_type should be SPIN or BINARY")


def measure_time(func):
    def wrapper(*args, **kargs):
        start_time = time.perf_counter()

        result = func(*args, **kargs)

        end_time = time.perf_counter()
        execution_time = end_time - start_time
        return execution_time, result
    return wrapper


def hubo_sa_sampling(bhom, var_type,
                     schedule, schedule_info,
                     num_sweeps=100, num_reads=1,
                     init_state=None, seed=None):

    variables = list(cast_var_type(var_type).value)
    init_state = init_state if init_state else np.random.choice(
        variables, len(bhom.indices))


    execution_time = []
    states, energies = [], []
    @measure_time
    def exec_sampling():
        for _ in range(num_reads):
            _exec_time, state = measure_time(
                hubo_simulated_annealing)(bhom, init_state, schedule,
                                          var_type=var_type)
            execution_time.append(_exec_time)
            states.append(state)
            energies.append(bhom.calc_energy(state))

    sampling_time, _ = exec_sampling()

    response = openjij.Response.from_samples(
        (states, bhom.indices), var_type, energies,
        info={}
    )

    response.info['sampling_time'] = sampling_time * 10**6  # micro sec
    response.info['execution_time'] = np.mean(
        execution_time) * 10**6  # micro sec
    response.info['list_exec_times'] = np.array(
        execution_time) * 10**6  # micro sec

    response.update_ising_states_energies(
        response.states, response.energies)

    response.info['schedule'] = schedule_info
    return response



import numpy as np
import openjij
from openjij.sampler import measure_time
from openjij.sampler import SQASampler
from openjij.utils.decorator import deprecated_alias
import cxxjij


class CSQASampler(SQASampler):
    """Sampler with continuous-time simulated quantum annealing (CSQA) using Hamiltonian

    .. math:: 

        H(s) = s H_p + \\Gamma (1-s)\\sum_i \\sigma_i^x

    where :math:`H_p` is the problem Hamiltonian we want to solve.

    Args:
        beta (float): Inverse temperature.
        gamma (float): Amplitude of quantum fluctuation.
        schedule (list): schedule list
        step_num (int): Number of Monte Carlo step.
        schedule_info (dict): Information about a annealing schedule.
        num_reads (int): Number of iterations.
        num_sweeps (int): number of sweeps
        num_reads (int): Number of iterations.
        schedule_info (dict): Information about a annealing schedule.

    """
    def __init__(self,
                 beta=5.0, gamma=1.0,
                 num_sweeps=1000, schedule=None,
                 num_reads=1):

        self.beta = beta
        self.gamma = gamma
        self.num_reads = num_reads
        self.num_sweeps = num_sweeps
        self.schedule = schedule
        self.energy_bias = 0.0
        self._schedule_setting = {
            'beta': beta,
            'gamma': gamma,
            'num_sweeps': num_sweeps,
            'num_reads': num_reads,
        }

    def _get_result(self, system, model):
        info = {}
        info['spin_config'] = system.spin_config

        state = cxxjij.result.get_solution(system)

        return state, info

    def sample_ising(self, h, J,
                     beta=None, gamma=None,
                     num_sweeps=None, schedule=None,
                     num_reads=1,
                     initial_state=None, updater='swendsenwang',
                     reinitialize_state=True, seed=None):
        """Sampling from the Ising model.

        Args:
            h (dict): linear biases
            J (dict): quadratic biases
            beta (float, optional): inverse temperature
            gamma (float, optional): strength of transverse field
            num_sweeps (int, optional): number of sampling.
            schedule (list, optional): schedule list
            num_reads (int, optional): number of iterations
            initial_state (optional): initial state of spins
            updater (str, optional): updater algorithm
            reinitialize_state (bool, optional): Re-initilization at each sampling. Defaults to True.
            seed (int, optional): Sampling seed.
            structure (int, optional): specify the structure. 
            This argument is necessary if the model has a specific structure (e.g. Chimera graph) and the updater algorithm is structure-dependent.
            structure must have two types of keys, namely "size" which shows the total size of spins and "dict" which is the map from model index (elements in model.indices) to the number.
        
        Returns:
            :class:`openjij.sampler.response.Response`: results

        Examples:
            
            for Ising case::

                >>> h = {0: -1, 1: -1, 2: 1, 3: 1}
                >>> J = {(0, 1): -1, (3, 4): -1}
                >>> sampler = oj.CSQASampler()
                >>> res = sampler.sample_ising(h, J)

            for QUBO case::

                >>> Q = {(0, 0): -1, (1, 1): -1, (2, 2): 1, (3, 3): 1, (4, 4): 1, (0, 1): -1, (3, 4): 1}
                >>> sampler = oj.CSQASampler()
                >>> res = sampler.sample_qubo(Q)

        """

        bqm = openjij.BinaryQuadraticModel(
            linear=h, quadratic=J, var_type='SPIN'
        )

        #Continuous time ising system only supports sparse ising graph

        ising_graph = bqm.get_cxxjij_ising_graph(sparse=True)

        self._setting_overwrite(
            beta=beta, gamma=gamma,
            num_sweeps=num_sweeps, num_reads=num_reads
        )
        self._annealing_schedule_setting(
            bqm, beta, gamma, num_sweeps, schedule)

        # make init state generator --------------------------------
        if initial_state is None:
            def init_generator():
                spin_config = np.random.choice([1,-1], len(bqm.indices))
                return list(spin_config)
        else:
            def init_generator(): return initial_state
        # -------------------------------- make init state generator

        # choose updater -------------------------------------------
        sqa_system = cxxjij.system.make_continuous_time_ising(
            init_generator(), ising_graph, self.gamma
        )
        _updater_name = updater.lower().replace('_', '').replace(' ', '')
        if _updater_name == 'swendsenwang':
            algorithm = cxxjij.algorithm.Algorithm_ContinuousTimeSwendsenWang_run
        else:
            raise ValueError('updater is one of "swendsen wang"')
        # ------------------------------------------- choose updater

        response = self._cxxjij_sampling(
            bqm, init_generator,
            algorithm, sqa_system,
            reinitialize_state, seed
        )

        response.info['schedule'] = self.schedule_info

        return response



import unittest
import openjij as oj

class HUBOTest(unittest.TestCase):
    def test_hubo_sampling(self):
        sampler = oj.SASampler()

        # make HUBO
        h = {0: -1}
        J = {(0, 1): -1}
        K = {(0, 1, 2): 1}

        response = sampler.sample_hubo([h, J, K], var_type="SPIN")



import numpy as np


class BinaryHigherOrderModel:
    """Higher order model.
    """

    def __init__(self, interactions: list):
        self.interactions = interactions

        indices = set(self.interactions[0].keys())
        for coeff in self.interactions[1:]:
            for _inds in coeff.keys():
                indices = indices | set(_inds)

        self.indices = list(indices)

        for i in self.indices:
            if i not in self.interactions[0]:
                self.interactions[0][i] = 0.0

    def adj_dict(self):
        """adjacency list of each variables

        Returns:
            dict: key (variables key), value (list of tuple represents connected indices)
        """
        adj_dict = {i: [] for i in self.indices}
        for coeff in self.interactions[1:]:
            for _inds, value in coeff.items():
                for i in _inds:
                    _inds_list = list(_inds)
                    _inds_list.remove(i)
                    adj_dict[i].append([_inds_list, value])
        return adj_dict

    def energy(self, state):
        """calculate energy of state

        Args:
            state (list of int): list of SPIN or BINARY 

        Returns:
            float: energy of state
        """
        energy = 0.0
        if isinstance(state, dict):
            # convert to array
            state = [state[elem] for elem in self.indices]

        state = np.array(state)
        for coeff in self.interactions[1:]:
            for _inds, value in coeff.items():
                energy += value * np.prod(state[list(_inds)])
        for i, hi in self.interactions[0].items():
            energy += hi * state[i]

        return energy

    def calc_energy(self, state):
        """alias of `energy`

        Args:
            state (list of int): list of SPIN or BINARY 

        Returns:
            float: energy of state
        """
        return self.energy(state)



from .response import Response
from .sampler import *
from .sa_sampler import SASampler
from .sqa_sampler import SQASampler
from .chimera_gpu import GPUChimeraSQASampler, GPUChimeraSASampler
from .csqa_sampler import CSQASampler
from .cmos_annealer import *



# Copyright 2019 Jij Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
| This module defines the BinaryQuadraticModel with the Hamiltonian,

.. math:: 

    H = \\sum_{i\\neq j} J_{ij}\sigma_i \sigma_j + \\sum_{i} h_{i}\sigma_i,

| in an Ising form and

.. math:: 

    H = \\sum_{ij} Q_{ij}x_i x_j + \\sum_{i} H_{i}x_i,

| in a QUBO form.
| The methods and usage are basically the same as `dimod <https://github.com/dwavesystems/dimod>`_.

"""

import numpy as np
import cxxjij
import openjij
from openjij.utils.graph_utils import qubo_to_ising
import cimod
import dimod
import warnings
import sys

def make_BinaryQuadraticModel(linear: dict, quadratic: dict):
    """BinaryQuadraticModel factory.

    Returns:
        generated BinaryQuadraticModel class
    """

    class BinaryQuadraticModel(cimod.make_BinaryQuadraticModel(linear, quadratic)):
        """Represents Binary quadratic model. 
           Indices are listed in self.indices.
        Attributes:
            var_type (dimod.Vartype): variable type SPIN or BINARY
            linear (dict): represents linear term
            quadratic (dict): represents quadratic term
            adj (dict): represents adjacency
            indices (list): labels of each variables sorted by results variables
            offset (float): represents constant energy term when convert to SPIN from BINARY
        """
    
        def __init__(self, linear: dict, quadratic: dict, offset: float=0.0,
                var_type=openjij.SPIN, gpu: bool=False, **kwargs):
            """BinaryQuadraticModel constructor.

            Args:
                linear (dict): linear biases.
                quadratic (dict): quadratic biases
                offset (float): offset
                var_type (openjij.variable_type.Vartype): var_type
                gpu (bool): if true, this can be used for gpu samplers.
                kwargs:
            """
    
            super().__init__(linear, quadratic, offset, var_type, **kwargs)
            self.gpu = gpu
    
    
        def get_cxxjij_ising_graph(self, sparse: bool=False):
            """generate cxxjij Ising graph from the interactions.

            Args:
                sparse (bool): if true, this function returns cxxjij.graph.Sparse. Otherwise it returns cxxjij.graph.Dense.
            Returns:
                cxxjij.graph.Dense or cxxjij.graph.Sparse: 
            """
    
            if sparse:
                GraphClass = cxxjij.graph.Sparse if self.gpu == False else cxxjij.graph.SparseGPU
                return GraphClass(self.to_serializable())
            else:
                GraphClass = cxxjij.graph.Dense if self.gpu == False else cxxjij.graph.DenseGPU
                # initialize with interaction matrix.
                mat = self.interaction_matrix()

                if (self.vartype == openjij.BINARY):
                    # convert to ising matrix
                    qubo_to_ising(mat)

                size = mat.shape[0]
                dense = GraphClass(size)
                # graph type is dense.
                # reshape matrix
                temp = np.zeros((size+1, size+1))
                temp[:size, :size] = mat
                temp[:size, size] = np.diag(mat)
                temp[size, :size] = np.diag(mat)
                np.fill_diagonal(temp, 0)
                temp[size, size] = 1
    
                dense.set_interaction_matrix(temp)
    
                return dense
    
    
        # compatible with the previous version
        def calc_energy(self, sample, **kwargs):
            return self.energy(sample, **kwargs)

    return BinaryQuadraticModel

def make_BinaryQuadraticModel_from_JSON(obj: dict):
    """make BinaryQuadraticModel from JSON.

    Returns:
        corresponding BinaryQuadraticModel type
    """

    label = obj['variable_labels'][0]
    if isinstance(label, list):
        #convert to tuple
        label = tuple(label)

    mock_linear = {label:1.0}

    return make_BinaryQuadraticModel(mock_linear, {})

def BinaryQuadraticModel(linear: dict, quadratic: dict, offset: float=0.0,
        var_type=dimod.SPIN, gpu: bool=False, **kwargs):
    """generate BinaryQuadraticModel object.

    Attributes:
        var_type (dimod.Vartype): variable type SPIN or BINARY
        linear (dict): represents linear term
        quadratic (dict): represents quadratic term
        adj (dict): represents adjacency
        indices (list): labels of each variables sorted by results variables.
        offset (float): represents constant energy term when convert to SPIN from BINARY
    Args:
        linear (dict): linear biases
        quadratic (dict): quadratic biases
        offset (float): offset
        var_type (openjij.variable_type.Vartype): vartype ('SPIN' or 'BINARY')
        gpu (bool): if true, this can be used for gpu samplers.
        kwargs:
    Returns:
        generated BinaryQuadraticModel
    Examples:
        BinaryQuadraticModel can be initialized by specifing h and J::

            >>> h = {0: 1, 1: -2}
            >>> J = {(0, 1): -1, (1, 2): -3, (2, 3): 0.5}
            >>> bqm = oj.BinaryQuadraticModel(self.h, self.J)

        You can also use strings and tuples of integers (up to 4 elements) as indices::

            >>> h = {'a': 1, 'b': -2}
            >>> J = {('a', 'b'): -1, ('b', 'c'): -3, ('c', 'd'): 0.5}
            >>> bqm = oj.BinaryQuadraticModel(self.h, self.J)
    """

    Model = make_BinaryQuadraticModel(linear, quadratic)

    return Model(linear, quadratic, offset, var_type, gpu, **kwargs)

#classmethods
BinaryQuadraticModel.from_qubo = \
        lambda Q, offset=0.0, **kwargs: make_BinaryQuadraticModel({}, Q).from_qubo(Q, offset, **kwargs)

BinaryQuadraticModel.from_ising = \
        lambda linear, quadratic, offset=0.0, **kwargs: make_BinaryQuadraticModel(linear, quadratic).from_ising(linear, quadratic, offset, **kwargs)

BinaryQuadraticModel.from_serializable = \
        lambda obj: make_BinaryQuadraticModel_from_JSON(obj).from_serializable(obj)



















import numpy as np
import openjij
from openjij.sampler import measure_time
from openjij.sampler import BaseSampler
from openjij.utils.decorator import deprecated_alias
import cxxjij
class SQASampler(BaseSampler):
    """Sampler with Simulated Quantum Annealing (SQA).

    Inherits from :class:`openjij.sampler.sampler.BaseSampler`.
    Hamiltonian

    .. math:: 

        H(s) = s H_p + \\Gamma (1-s)\\sum_i \\sigma_i^x

    where :math:`H_p` is the problem Hamiltonian we want to solve.

    Args:
        beta (float): Inverse temperature.
        gamma (float): Amplitude of quantum fluctuation.
        trotter (int): Trotter number.
        num_sweeps (int): number of sweeps
        schedule (list): schedule list
        num_reads (int): Number of iterations.
        schedule_info (dict): Information about a annealing schedule.

    Raises:
        ValueError: If the schedule violates as below.
        - not list or numpy.array.
        - schedule range is '0 <= s <= 1'.

    """
    @property
    def parameters(self):
        return {
            'beta': ['parameters'],
            'gamma': ['parameters'],
            'trotter': ['parameters'],
        }

    @deprecated_alias(iteration='num_reads')
    def __init__(self,
                 beta=5.0, gamma=1.0,
                 num_sweeps=1000, schedule=None,
                 trotter=4,
                 num_reads=1):

        self.beta = beta
        self.gamma = gamma
        self.trotter = trotter
        self.num_reads = num_reads
        self.num_sweeps = num_sweeps
        self.schedule = schedule
        self._schedule_setting = {
            'beta': beta,
            'gamma': gamma,
            'num_sweeps': num_sweeps,
            'num_reads': num_reads,
        }

        self._make_system = {
            'singlespinflip': cxxjij.system.make_transverse_ising
        }
        self._algorithm = {
            'singlespinflip': cxxjij.algorithm.Algorithm_SingleSpinFlip_run
        }

    def _convert_validation_schedule(self, schedule, beta):
        if not isinstance(schedule, (list, np.array)):
            raise ValueError("schedule should be list or numpy.array")

        if isinstance(schedule[0], cxxjij.utility.TransverseFieldSchedule):
            return schedule

        # schedule validation  0 <= s <= 1
        sch = np.array(schedule).T[0]
        if not np.all((0 <= sch) & (sch <= 1)):
            raise ValueError("schedule range is '0 <= s <= 1'.")

        if len(schedule[0]) == 2:
            # schedule element: (s, one_mc_step) with beta fixed
            # convert to list of cxxjij.utility.TransverseFieldSchedule
            cxxjij_schedule = []
            for s, one_mc_step in schedule:
                _schedule = cxxjij.utility.TransverseFieldSchedule()
                _schedule.one_mc_step = one_mc_step
                _schedule.updater_parameter.beta = beta
                _schedule.updater_parameter.s = s
                cxxjij_schedule.append(_schedule)
            return cxxjij_schedule
        elif len(schedule[0]) == 3:
            # schedule element: (s, beta, one_mc_step)
            # convert to list of cxxjij.utility.TransverseFieldSchedule
            cxxjij_schedule = []
            for s, _beta, one_mc_step in schedule:
                _schedule = cxxjij.utility.TransverseFieldSchedule()
                _schedule.one_mc_step = one_mc_step
                _schedule.updater_parameter.beta = _beta
                _schedule.updater_parameter.s = s
                cxxjij_schedule.append(_schedule)
            return cxxjij_schedule
        else:
            raise ValueError(
                """schedule is list of tuple or list
                (annealing parameter s : float, step_length : int) or
                (annealing parameter s : float, beta: float, step_length : int)
                """)

    def _get_result(self, system, model):
        state, info = super()._get_result(system, model)

        q_state = system.trotter_spins[:-1].T.astype(np.int)
        c_energies = [model.energy(
            state, convert_sample=True) for state in q_state]
        info['trotter_state'] = q_state
        info['trotter_energies'] = c_energies

        return state, info

    def sample_ising(self, h, J,
                     beta=None, gamma=None,
                     num_sweeps=None, schedule=None, trotter=None,
                     num_reads=1,
                     initial_state=None, updater='single spin flip',
                     reinitialize_state=True, seed=None, structure=None):
        """Sampling from the Ising model

        Args:
            h (dict): Linear term of the target Ising model. 
            J (dict): Quadratic term of the target Ising model. 
            beta (float, optional): inverse tempareture.
            gamma (float, optional): strangth of transverse field. Defaults to None.
            num_sweeps (int, optional): number of sweeps. Defaults to None.
            schedule (list[list[float, int]], optional): List of annealing parameter. Defaults to None.
            trotter (int): Trotter number.
            num_reads (int, optional): number of sampling. Defaults to 1.
            initial_state (list[int], optional): Initial state. Defaults to None.
            updater (str, optional): update method. Defaults to 'single spin flip'.
            reinitialize_state (bool, optional): Re-initilization at each sampling. Defaults to True.
            seed (int, optional): Sampling seed. Defaults to None.
            structure (dict): specify the structure. 
            This argument is necessary if the model has a specific structure (e.g. Chimera graph) and the updater algorithm is structure-dependent.
            structure must have two types of keys, namely "size" which shows the total size of spins and "dict" which is the map from model index (elements in model.indices) to the number.

        Raises:
            ValueError: 

        Returns:
            :class:`openjij.sampler.response.Response`: results

        Examples:
            
            for Ising case::

                >>> h = {0: -1, 1: -1, 2: 1, 3: 1}
                >>> J = {(0, 1): -1, (3, 4): -1}
                >>> sampler = oj.SQASampler()
                >>> res = sampler.sample_ising(h, J)

            for QUBO case::

                >>> Q = {(0, 0): -1, (1, 1): -1, (2, 2): 1, (3, 3): 1, (4, 4): 1, (0, 1): -1, (3, 4): 1}
                >>> sampler = oj.SQASampler()
                >>> res = sampler.sample_qubo(Q)
        """

        bqm = openjij.BinaryQuadraticModel(
            linear=h, quadratic=J, var_type='SPIN'
        )
        return self._sampling(bqm, beta=beta, gamma=gamma,
                     num_sweeps=num_sweeps, schedule=schedule, trotter=trotter,
                     num_reads=num_reads,
                     initial_state=initial_state, updater=updater,
                     reinitialize_state=reinitialize_state, seed=seed, structure=structure)

    def _sampling(self, bqm, beta=None, gamma=None,
                     num_sweeps=None, schedule=None, trotter=None,
                     num_reads=1,
                     initial_state=None, updater='single spin flip',
                     reinitialize_state=True, seed=None, structure=None):

        ising_graph = bqm.get_cxxjij_ising_graph()

        self._setting_overwrite(
            beta=beta, gamma=gamma,
            num_sweeps=num_sweeps, num_reads=num_reads,
            trotter=trotter
        )

        # set annealing schedule -------------------------------
        self._annealing_schedule_setting(
            bqm, beta, gamma, num_sweeps, schedule)
        # ------------------------------- set annealing schedule

        # make init state generator --------------------------------
        if initial_state is None:
            def init_generator(): return [ising_graph.gen_spin(seed) if seed != None else ising_graph.gen_spin()
                                          for _ in range(self.trotter)]
        else:
            if isinstance(initial_state, dict):
                initial_state = [initial_state[k] for k in bqm.indices]
            _init_state = np.array(initial_state)

            if structure == None:
                # validate initial_state size
                if len(initial_state) != ising_graph.size():
                    raise ValueError(
                        "the size of the initial state should be {}"
                        .format(ising_graph.size()))
            else:
                # resize _initial_state
                temp_state = [1]*int(structure['size'])
                for k,ind in enumerate(model.indices):
                    temp_state[structure['dict'][ind]] = _init_state[k]
                _init_state = temp_state

            trotter_init_state = [_init_state
                                  for _ in range(self.trotter)]

            def init_generator(): return trotter_init_state
        # -------------------------------- make init state generator

        # choose updater -------------------------------------------
        _updater_name = updater.lower().replace('_', '').replace(' ', '')
        if _updater_name not in self._algorithm:
            raise ValueError('updater is one of "single spin flip"')
        algorithm = self._algorithm[_updater_name] 
        sqa_system = self._make_system[_updater_name](
            init_generator(), ising_graph, self.gamma
        )
        # ------------------------------------------- choose updater

        response = self._cxxjij_sampling(
            bqm, init_generator,
            algorithm, sqa_system,
            reinitialize_state, seed, structure
        )

        response.info['schedule'] = self.schedule_info

        return response

    def _annealing_schedule_setting(self, model,
                                    beta=None, gamma=None,
                                    num_sweeps=None,
                                    schedule=None):
        self.beta = beta if beta else self.beta
        self.gamma = gamma if gamma else self.gamma
        if schedule or self.schedule:
            self._schedule = self._convert_validation_schedule(
                schedule if schedule else self.schedule, self.beta
            )
            self.schedule_info = {'schedule': 'custom schedule'}
        else:

            self.num_sweeps = num_sweeps if num_sweeps else self.num_sweeps
            self._schedule, beta_gamma = quartic_ising_schedule(
                model=model,
                beta=self._schedule_setting['beta'],
                gamma=self._schedule_setting['gamma'],
                num_sweeps=self._schedule_setting['num_sweeps']
            )
            self.schedule_info = {
                'beta': beta_gamma[0],
                'gamma': beta_gamma[1],
                'num_sweeps': self._schedule_setting['num_sweeps']
            }


def linear_ising_schedule(model, beta, gamma, num_sweeps):
    """Generate linear ising schedule.

    Args:
        model (:class:`openjij.model.model.BinaryQuadraticModel`): BinaryQuadraticModel
        beta (float): inverse temperature
        gamma (float): transverse field
        num_sweeps (int): number of steps
    Returns:
        generated schedule
    """
    schedule = cxxjij.utility.make_transverse_field_schedule_list(
        beta=beta, one_mc_step=1, num_call_updater=num_sweeps
    )
    gamma = 1
    return schedule, [beta, gamma]

#TODO: more optimal schedule?
def quartic_ising_schedule(model, beta, gamma, num_sweeps):
    """Generate quartic ising schedule based on S. Morita and H. Nishimori, Journal of Mathematical Physics 49, 125210 (2008).

    Args:
        model (:class:`openjij.model.model.BinaryQuadraticModel`): BinaryQuadraticModel
        beta (float): inverse temperature
        gamma (float): transverse field
        num_sweeps (int): number of steps
    Returns:
        generated schedule
    """

    s = np.linspace(0, 1, num_sweeps)
    fs = s**4*(35-84*s+70*s**2-20*s**3)
    schedule = [((beta, elem), 1) for elem in fs]
    gamma = 1
    return schedule, [beta, gamma]




import unittest

import numpy as np
import openjij as oj
import cxxjij as cj


def calculate_ising_energy(h, J, spins):
    energy = 0.0
    for (i, j), Jij in J.items():
        energy += Jij*spins[i]*spins[j]
    for i, hi in h.items():
        energy += hi * spins[i]
    return energy


def calculate_qubo_energy(Q, binary):
    energy = 0.0
    for (i, j), Qij in Q.items():
        energy += Qij*binary[i]*binary[j]
    return energy


class VariableTypeTest(unittest.TestCase):
    def test_variable_type(self):
        spin = oj.cast_var_type('SPIN')
        self.assertEqual(spin, oj.SPIN)

        binary = oj.cast_var_type('BINARY')
        self.assertEqual(binary, oj.BINARY)


class ModelTest(unittest.TestCase):

    def setUp(self):
        self.h = {0: 1, 1: -2}
        self.J = {(0, 1): -1, (1, 2): -3, (2, 3): 0.5}
        self.spins = {0: 1, 1: -1, 2: 1, 3: 1}

        self.Q = {(0, 0): 1, (1, 2): -1, (2, 0): -0.2, (1, 3): 3}
        self.binaries = {0: 0, 1: 1, 2: 1, 3: 0}

    def test_bqm_constructor(self):
        # Test BinaryQuadraticModel constructor
        bqm = oj.BinaryQuadraticModel(self.h, self.J)
        self.assertEqual(type(bqm.interaction_matrix()), np.ndarray)

        self.assertEqual(bqm.vartype, oj.SPIN)

        dense_graph = bqm.get_cxxjij_ising_graph(sparse=False)
        self.assertTrue(isinstance(dense_graph, cj.graph.Dense))

        bqm_qubo = oj.BinaryQuadraticModel.from_qubo(Q=self.Q)
        self.assertEqual(bqm_qubo.vartype, oj.BINARY)

    def test_interaction_matrix(self):
        bqm = oj.BinaryQuadraticModel(self.h, self.J)
        ising_matrix = np.array([
            [1, -1,  0,  0],
            [-1, -2, -3, 0],
            [0, -3, 0, 0.5],
            [0, 0, 0.5, 0]
        ])
        np.testing.assert_array_equal(
            bqm.interaction_matrix(), ising_matrix
        )

        # check Hij = Jij + Jji
        J = self.J.copy()
        J[0, 1] /= 3
        J[1, 0] = J[0, 1] * 2
        bqm = oj.BinaryQuadraticModel(self.h, J)
        np.testing.assert_array_equal(bqm.interaction_matrix(), ising_matrix)

    def test_transfer_to_cxxjij(self):
        bqm = oj.BinaryQuadraticModel(self.h, self.J)
        # to Dense
        ising_graph = bqm.get_cxxjij_ising_graph(sparse=False)
        self.assertEqual(ising_graph.size(), len(bqm.indices))
        for i in range(len(bqm.indices)):
            for j in range(len(bqm.indices)):
                if i != j:
                    self.assertAlmostEqual(bqm.interaction_matrix()[i,j], ising_graph.get_interactions()[i, j])
                else:
                    # i == j
                    self.assertAlmostEqual(bqm.interaction_matrix()[i,j], ising_graph.get_interactions()[i, len(bqm.indices)])
                    self.assertAlmostEqual(bqm.interaction_matrix()[i,j], ising_graph.get_interactions()[len(bqm.indices), i])
                    self.assertEqual(ising_graph.get_interactions()[i,i], 0)

        self.assertEqual(ising_graph.get_interactions()[len(bqm.indices),len(bqm.indices)], 1)


        # to Sparse
        ising_graph = bqm.get_cxxjij_ising_graph(sparse=True)
        self.assertEqual(ising_graph.size(), len(bqm.indices))
        for i in range(ising_graph.size()):
            for j in ising_graph.adj_nodes(i):
                self.assertEqual(bqm.interaction_matrix()[i,j], ising_graph[i,j])


    def test_bqm_calc_energy(self):
        # Test to calculate energy

        # Test Ising energy
        bqm = oj.BinaryQuadraticModel(self.h, self.J)
        ising_energy_bqm = bqm.energy(self.spins)
        true_ising_e = calculate_ising_energy(self.h, self.J, self.spins)
        self.assertEqual(ising_energy_bqm, true_ising_e)

        # Test QUBO energy
        bqm = oj.BinaryQuadraticModel.from_qubo(Q=self.Q)
        qubo_energy_bqm = bqm.energy(self.binaries)
        true_qubo_e = calculate_qubo_energy(self.Q, self.binaries)
        self.assertEqual(qubo_energy_bqm, true_qubo_e)

        # QUBO == Ising
        spins = {0: 1, 1: 1, 2: -1, 3: 1}
        binary = {0: 1, 1: 1, 2: 0, 3: 1}
        qubo_bqm = oj.BinaryQuadraticModel.from_qubo(Q=self.Q)
        # ising_mat = qubo_bqm.ising_interactions()
        # h, J = {}, {}
        # for i in range(len(ising_mat)-1):
        #     for j in range(i, len(ising_mat)):
        #         if i == j:
        #             h[i] = ising_mat[i][i]
        #         else:
        #             J[(i, j)] = ising_mat[i][j]

        qubo_energy = qubo_bqm.energy(binary)

        self.assertEqual(qubo_energy, qubo_bqm.energy(spins, convert_sample=True))

    def test_energy_consistency(self):
        bqm = oj.BinaryQuadraticModel(self.h, self.J, var_type='SPIN')
        dense_ising_graph = bqm.get_cxxjij_ising_graph(sparse=False)
        sparse_ising_graph = bqm.get_cxxjij_ising_graph(sparse=True)
        spins = {0: -1, 1: -1, 2: -1, 3: -1}
        self.assertAlmostEqual(dense_ising_graph.calc_energy([spins[i] for i in range(len(spins))]), bqm.energy(spins))
        self.assertAlmostEqual(sparse_ising_graph.calc_energy([spins[i] for i in range(len(spins))]), bqm.energy(spins))

    def test_bqm(self):
        h = {}
        J = {(0, 1): -1.0, (1, 2): -3.0}
        bqm = oj.BinaryQuadraticModel(h, J)
        
        self.assertEqual(J, bqm.get_quadratic())

        self.assertEqual(type(bqm.interaction_matrix()), np.ndarray)
        correct_mat = np.array([[0, -1, 0, ], [-1, 0, -3], [0, -3, 0]])
        np.testing.assert_array_equal(
            bqm.interaction_matrix(), correct_mat.astype(np.float))

    def test_chimera_converter(self):
        h = {}
        J = {(0, 4): -1.0, (6, 2): -3.0, (16, 0): 4}
        chimera = oj.ChimeraModel(h, J, offset=0, unit_num_L=2)
        self.assertEqual(chimera.chimera_coordinate(
            4, unit_num_L=2), (0, 0, 4))
        self.assertEqual(chimera.chimera_coordinate(
            12, unit_num_L=2), (0, 1, 4))
        self.assertEqual(chimera.chimera_coordinate(
            16, unit_num_L=2), (1, 0, 0))

    def test_chimera(self):
        h = {}
        J = {(0, 4): -1.0, (6, 2): -3.0}
        bqm = oj.ChimeraModel(h, J, offset=0, unit_num_L=3)
        self.assertTrue(bqm.validate_chimera())

        J = {(0, 1): -1}
        bqm = oj.ChimeraModel(h, J, unit_num_L=3)
        with self.assertRaises(ValueError):
            bqm.validate_chimera()

        J = {(4, 12): -1}
        bqm = oj.ChimeraModel(h, J, unit_num_L=2)
        self.assertTrue(bqm.validate_chimera())

        J = {(0, 4): -1, (5, 13): 1, (24, 8): 2,
             (18, 20): 1, (16, 0): 0.5, (19, 23): -2}
        h = {13: 2}
        chimera = oj.ChimeraModel(h, J, unit_num_L=2)
        self.assertEqual(chimera.to_index(1, 1, 1, unit_num_L=2), 25)

        self.assertTrue(chimera.validate_chimera())

    def test_ising_dict(self):
        Q = {(0, 4): -1.0, (6, 2): -3.0}
        bqm = oj.ChimeraModel.from_qubo(Q=Q, unit_num_L=3)

    def test_king_graph(self):
        h = {}
        J = {(0, 1): -1.0, (1, 2): -3.0}
        king_interaction = [[0, 0, 1, 0, -1.0], [1, 0, 2, 0, -3.0]]

        king_graph = oj.KingGraph(machine_type="ASIC", linear=h, quadratic=J)
        correct_mat = np.array([[0, -1, 0, ], [-1, 0, -3], [0, -3, 0]])
        np.testing.assert_array_equal(
            king_graph.interaction_matrix(), correct_mat.astype(np.float))

        self.assertCountEqual(king_interaction, king_graph._ising_king_graph)
    
        king_graph = oj.KingGraph(
            machine_type="ASIC", king_graph=king_interaction)
        
        np.testing.assert_array_equal(
            king_interaction, king_graph._ising_king_graph)

        king_graph = oj.KingGraph.from_qubo(Q={(0, 1): -1}, machine_type='ASIC')
        king_interaction = [[0, 0, 0, 0, -0.25],
                            [0, 0, 1, 0, -0.25], [1, 0, 1, 0, -0.25]]
        self.assertCountEqual(king_interaction, king_graph._ising_king_graph)
    
    def test_get_chimera_graph(self):
        c_model = oj.ChimeraModel.from_qubo(Q={(0, 4): -1, (1, 1): -1, (1, 5): 1}, unit_num_L=2)
        chimera = c_model.get_cxxjij_ising_graph()
        self.assertIsInstance(chimera, cj.graph.Chimera)

        c_model = oj.ChimeraModel.from_qubo(Q={((0, 0, 1), (0, 0, 4)): -1, ((0, 0, 4), (0, 0, 2)): -1}, 
                                            unit_num_L=2)
        chimera = c_model.get_cxxjij_ising_graph()
        self.assertIsInstance(chimera, cj.graph.Chimera)


if __name__ == '__main__':
    unittest.main()



# Copyright 2019 Jij Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests
import json

from openjij.sampler import BaseSampler
import openjij

import numpy as np
import dimod


class CMOSAnnealer(BaseSampler, dimod.Structured):
    """Sampler with CMOS Annealer.

    Inherits from :class:`openjij.sampler.sampler.BaseSampler`.

    To get More details about CMOS Annealer API,
    please access the reference (https://annealing-cloud.com/web-api/reference.html).

    Args:
        token (str): API token of COMS Annealer.
        machine_type (str): Type of CMOS Annealer: 'ASIC' or 'FPGA'.
        beta_min (float): Minimum beta (inverse temperature).
        beta_max (float): Maximum beta (inverse temperature).
        step_length (int): Length of Monte Carlo step.
        num_sweeps (int): Number of Monte Carlo step.
        num_reads (int): Number of iterations.
        **kwargs: Optional keyword arguments for CMOS Annealer.

    Attributes:
        cmos_parameters (dict): Parameters of CMOS Annealer.

    Raises:
        ValueError:
            - only valid input: h, J, Q, or king graph.
            - CMOS Annealer API raises error.

    """

    def __init__(self, token, machine_type="ASIC",
                 beta_min=None, beta_max=None,
                 num_sweeps=10, step_length=100, num_reads=1, **kwargs):

        self.token = token
        self.machine_type = machine_type
        # setting machine type and 2D Grid size
        # more information see: https://annealing-cloud.com/web-api/reference/v2.html
        if machine_type == 'ASIC':
            self._X, self._Y = 352, 176
            self.max_abs = 3
        elif machine_type == 'FPGA':
            self._X, self._Y = 80, 80
            self.max_abs = 127
        else:
            raise ValueError("machine_type is 'ASIC' or 'FPGA'.")

        self.beta_min = beta_min
        self.beta_max = beta_max
        self.num_sweeps = num_sweeps
        self.num_reads = num_reads
        self._schedule_setting = {
            'beta_min': beta_min,
            'beta_max': beta_max,
            'num_sweeps': num_sweeps,
            'num_reads': num_reads,
            'step_length': step_length
        }

    @property
    def parameters(self):
        return {'num_reads': ['1<=n<=10', lambda n: 1 <= n <= 10],
                'num_sweeps': ['1<=n<=100', lambda n: 1 <= n <= 100],
                'step_length': ['1<=n<=1000', lambda n: 1 <= n <= 1000],
                'beta_min': ['1/2000<=beta', lambda beta: 1/2000 <= beta],
                'beta_max': ['1/2000<=beta', lambda beta: 1/2000 <= beta]}

    @property
    def properties(self):
        return {'System size': "self._X, self._Y are size of 2D grid King's graph."}

    @property
    def nodelist(self):
        """2-D lattice nodes

        Returns:
            [list of int]: 2-D lattice nodes
        """
        return list(range(self._X*self._Y))

    @property
    def edgelist(self):
        """King's graph

        Returns:
            [list of tuple]: King's graph
        """
        edges = []
        for _x in range(self._X-1):
            for _y in range(self._Y-1):
                i = _x + self._X * _y
                edges += [
                    (i, _x + 1 + self._X * _y),
                    (i, _x + self._X * (_y + 1)),
                    (i, _x + 1 + self._X * (_y + 1)),
                ]
        return edges

    def _validate_schedule(self):
        for k, v in self._schedule_setting.items():
            if not self.parameters[k][1](v):
                raise ValueError(
                    'The value of {} is invalid.'
                    ' "{}: {}".'.format(k, k, self.parameters[k][0]))

    def sample_ising(self, h, J, num_reads=1,
                     num_sweeps=None, step_length=None,
                     beta_min=None, beta_max=None,):
        """sample Ising model.

        Args:
            h (dict): linear biases
            J (dict): quadratic biases
            num_reads (int): number of reads
            num_sweeps (int): number of sweeps
            step_length (int): number of Monte Carlo steop for each step
            beta_min (float): minimal value of inverse temperature
            beta_max (float): maximum value of inverse temperature
        """

        # settting annealing parameters and validation -------------
        self._setting_overwrite(
            beta_min=beta_min, beta_max=beta_max,
            num_sweeps=num_sweeps, num_reads=num_reads,
            step_length=step_length
        )
        # ------------- settting annealing parameters and validation

        # Convert to CMOS Annealing Cloud Web API format [xi,yi,xj,yj,pij] ---
        # from the Ising interaction dictionary.
        cmos_model = _cmos_annealer_format(
            h, J, self._X, self._Y, self.max_abs)
        beta_max, beta_min = defalut_beta_range(
            cmos_model,
            beta_max=self._schedule_setting['beta_max'],
            beta_min=self._schedule_setting['beta_min']
        )
        self._schedule_setting['beta_max'] = beta_max
        self._schedule_setting['beta_min'] = beta_min
        self._validate_schedule()
        headers, request = self.make_json_request(
            cmos_model, self.token
        )
        # --------------------------------------------------------------------

        # POST to API -------------------------------------
        url = 'https://annealing-cloud.com/api/v2/solve'
        res = requests.post(url, data=json.dumps(
            request, cls=NumpyEncoder), headers=headers)
        res_dict = res.json()
        # ------------------------------------- POST to API

        # API Error handling
        if res_dict['status'] != 0:
            raise ValueError('Error status: {}, message: {}'.format(
                res_dict['status'], res_dict['message']))

        # more infomation see : https://annealing-cloud.com/web-api/reference/v2.html
        info = {
            "schedule": {
                'beta_max': self._schedule_setting['beta_max'],
                'beta_min': self._schedule_setting['beta_min']
            },
            "averaged_spins": res_dict['result']["averaged_spins"],
            "averaged_energy": res_dict['result']["averaged_energy"],
            'execution_time': res_dict['result']['execution_time'] * 10**(-3),
            'job_id': res_dict['job_id'],
            'posted_model': cmos_model
        }

        bqm = openjij.BinaryQuadraticModel(h, J)
        samples = [{x+self._X*y: s for x, y, s in spins}
                   for spins in res_dict['result']['spins']]
        response = openjij.Response.from_samples_bqm(
            samples, bqm, info=info
        )
        return response

    def make_json_request(self, model, token):
        """Make request for CMOS Annealer API.

        Args:
            model (list):
                A list of 5 integer values representing vertices
                or interactions of Ising model

            token (str):
                API token of COMS Annealer.

        """

        headers = {"Authorization": "Bearer " + token}
        headers.update({"Accept": "application/json"})
        headers.update({'content-type': 'application/json'})

        request = {}
        request["model"] = model  # modelのみ必須項目
        request["type"] = 1 if self.machine_type == "ASIC" else 2  # FPGA
        request["num_executions"] = self.num_reads
        request["parameter"] = {
            "temperature_num_steps": self._schedule_setting['num_sweeps'],
            "temperature_step_length": self._schedule_setting['step_length'],
            "temperature_initial": 1.0/self._schedule_setting['beta_min'],
            "temperature_target": 1.0/self._schedule_setting['beta_max']}
        request["outputs"] = {"energies": True,
                              "spins": True,
                              "execution_time": True,
                              "num_outputs": 0,
                              "averaged_spins": True,
                              "averaged_energy": True}

        return headers, request


def _cmos_annealer_format(h, J, X, Y, max_abs):

    interactions = []

    def _2d_pos(i, j, pij):
        xi, xj = i % X, j % X
        yi, yj = int((i-xi)/X), int((j-xj)/X)
        interactions.append(pij)
        return [xi, yi, xj, yj, pij]

    annealer_format = [_2d_pos(i, j, Jij) for (i, j), Jij in J.items()]
    annealer_format += [_2d_pos(i, i, hi) for i, hi in h.items()]
    # degitalize to integer
    max_int = np.max(np.abs(interactions))
    _interactions = np.array(annealer_format).T
    _interactions[-1] = _interactions[-1] * max_abs/max_int
    _interactions = [list(q) for q in _interactions.T.astype(int)]
    return _interactions


def defalut_beta_range(cmos_model, beta_max, beta_min):

    if beta_min is None or beta_max is None:
        interactions = {}

        def add_interactions(i, v):
            if i in interactions:
                interactions[i].append(np.abs(v))
            else:
                interactions[i] = [np.abs(v)]

        for xi, yi, xj, yj, Jij in cmos_model:
            add_interactions((xi, yi), Jij)
            add_interactions((xj, yj), Jij)

        abs_bias = np.array([np.sum(v) for v in interactions.values()])
        ising_interaction = np.abs(np.array(cmos_model).T[-1])
        min_delta_energy = np.min(ising_interaction[ising_interaction > 0])
        max_delta_energy = np.max(abs_bias[abs_bias > 0])

    beta_min = np.log(
        2) / max_delta_energy if beta_min is None else beta_min
    beta_max = np.log(
        100) / min_delta_energy if beta_max is None else beta_max

    beta_min = max(beta_min, 1/2000)
    beta_max = max(beta_max, 1/2000)

    return beta_max, beta_min


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NumpyEncoder, self).default(obj)



# Copyright 2019 Jij Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License

import numpy as np
import openjij
from openjij.sampler import BaseSampler
from openjij.utils.decorator import deprecated_alias
from openjij.utils.graph_utils import qubo_to_ising
from openjij.model import BinaryHigherOrderModel
from .hubo_simulated_annealing import hubo_sa_sampling
from .hubo_simulated_annealing import default_schedule
import cxxjij

"""
This module contains Simulated Annealing sampler.
"""

class SASampler(BaseSampler):
    """Sampler with Simulated Annealing (SA).

    Args:
        beta_min (float):
            Minmum beta (inverse temperature).
            You can overwrite in methods .sample_*.

        beta_max (float):
            Maximum beta (inverse temperature).
            You can overwrite in methods .sample_*.

        num_reads (int):
            number of sampling (algorithm) runs. defaults None.
            You can overwrite in methods .sample_*.

        num_sweeps (int):
            number of MonteCarlo steps during SA. defaults None.
            You can overwrite in methods .sample_*.

        schedule_info (dict):
            Information about an annealing schedule.

    Raises:
        ValueError: If schedules or variables violate as below.
        - not list or numpy.array.
        - not list of tuple (beta : float, step_length : int).
        - beta is less than zero.

    """

    @property
    def parameters(self):
        return {
            'beta_min': ['parameters'],
            'beta_max': ['parameters'],
        }

    def __init__(self,
                 beta_min=None, beta_max=None,
                 num_sweeps=1000, schedule=None,
                 num_reads=1):

        self.beta_min = beta_min
        self.beta_max = beta_max
        self.num_sweeps = num_sweeps
        self.schedule = schedule
        self.num_reads = num_reads
        self._schedule_setting = {
            'beta_min': beta_min,
            'beta_max': beta_max,
            'num_sweeps': num_sweeps,
            'num_reads': num_reads,
        }

        self._make_system = {
            'singlespinflip': cxxjij.system.make_classical_ising,
            'swendsenwang': cxxjij.system.make_classical_ising
        }
        self._algorithm = {
            'singlespinflip': cxxjij.algorithm.Algorithm_SingleSpinFlip_run,
            'swendsenwang': cxxjij.algorithm.Algorithm_SwendsenWang_run
        }


    def _convert_validation_schedule(self, schedule):
        """Checks if the schedule is valid and returns cxxjij schedule
        """
        if not isinstance(schedule, (list, np.array)):
            raise ValueError("schedule should be list or numpy.array")

        if isinstance(schedule[0], cxxjij.utility.ClassicalSchedule):
            return schedule

        if len(schedule[0]) != 2:
            raise ValueError(
                "schedule is list of tuple or list (beta : float, step_length : int)")

        # schedule validation  0 <= beta
        beta = np.array(schedule).T[0]
        if not np.all(0 <= beta):
            raise ValueError("schedule beta range is '0 <= beta'.")

        # convert to cxxjij.utility.ClassicalSchedule
        cxxjij_schedule = []
        for beta, step_length in schedule:
            _schedule = cxxjij.utility.ClassicalSchedule()
            _schedule.one_mc_step = step_length
            _schedule.updater_parameter.beta = beta
            cxxjij_schedule.append(_schedule)

        return cxxjij_schedule

    def sample_ising(self, h, J, beta_min=None, beta_max=None,
                     num_sweeps=None, num_reads=1, schedule=None,
                     initial_state=None, updater='single spin flip',
                     reinitialize_state=True, seed=None,
                     ):
        """sample Ising model.

        Args:
            h (dict): linear biases
            J (dict): quadratic biases
            beta_min (float): minimal value of inverse temperature
            beta_max (float): maximum value of inverse temperature
            num_sweeps (int): number of sweeps
            num_reads (int): number of reads
            schedule (list): list of inverse temperature
            initial_state (dict): initial state
            updater(str): updater algorithm
            reinitialize_state (bool): if true reinitialize state for each run
            seed (int): seed for Monte Carlo algorithm
        Returns:
            :class:`openjij.sampler.response.Response`: results
            
        Examples:
            
            for Ising case::

                >>> h = {0: -1, 1: -1, 2: 1, 3: 1}
                >>> J = {(0, 1): -1, (3, 4): -1}
                >>> sampler = oj.SASampler()
                >>> res = sampler.sample_ising(h, J)

            for QUBO case::

                >>> Q = {(0, 0): -1, (1, 1): -1, (2, 2): 1, (3, 3): 1, (4, 4): 1, (0, 1): -1, (3, 4): 1}
                >>> sampler = oj.SASampler()
                >>> res = sampler.sample_qubo(Q)
            
        """

        model = openjij.BinaryQuadraticModel(
            linear=h, quadratic=J, var_type='SPIN'
        )
        return self._sampling(model, beta_min, beta_max,
                              num_sweeps, num_reads, schedule,
                              initial_state, updater,
                              reinitialize_state, seed)

    def _sampling(self, model, beta_min=None, beta_max=None,
                     num_sweeps=None, num_reads=1, schedule=None,
                     initial_state=None, updater='single spin flip',
                     reinitialize_state=True, seed=None, structure=None, 
                     ):
        """sampling by using specified model
        Args:
            model (openjij.BinaryQuadraticModel): BinaryQuadraticModel
            beta_min (float): minimal value of inverse temperature
            beta_max (float): maximum value of inverse temperature
            num_sweeps (int): number of sweeps
            num_reads (int): number of reads
            schedule (list): list of inverse temperature
            initial_state (dict): initial state
            updater(str): updater algorithm
            reinitialize_state (bool): if true reinitialize state for each run
            seed (int): seed for Monte Carlo algorithm
            structure (dict): specify the structure. 
            This argument is necessary if the model has a specific structure (e.g. Chimera graph) and the updater algorithm is structure-dependent.
            structure must have two types of keys, namely "size" which shows the total size of spins and "dict" which is the map from model index (elements in model.indices) to the number.
        Returns:
            :class:`openjij.sampler.response.Response`: results
        """
        _updater_name = updater.lower().replace('_', '').replace(' ', '')
        # swendsen wang algorithm runs only on sparse ising graphs.
        if _updater_name == 'swendsenwang':
            ising_graph = model.get_cxxjij_ising_graph(sparse=True)
        else:
            ising_graph = model.get_cxxjij_ising_graph()


        self._setting_overwrite(
            beta_min=beta_min, beta_max=beta_max,
            num_sweeps=num_sweeps, num_reads=num_reads
        )


        # set annealing schedule -------------------------------
        if schedule or self.schedule:
            self._schedule = self._convert_validation_schedule(
                schedule if schedule else self.schedule
            )
            self.schedule_info = {'schedule': 'custom schedule'}
        else:
            self._schedule, beta_range = geometric_ising_beta_schedule(
                model=model,
                beta_max=self._schedule_setting['beta_max'],
                beta_min=self._schedule_setting['beta_min'],
                num_sweeps=self._schedule_setting['num_sweeps']
            )
            self.schedule_info = {
                'beta_max': beta_range[0],
                'beta_min': beta_range[1],
                'num_sweeps': self._schedule_setting['num_sweeps']
            }
        # ------------------------------- set annealing schedule

        # make init state generator --------------------------------
        if initial_state is None:
            def _generate_init_state(): return ising_graph.gen_spin(seed) if seed != None else ising_graph.gen_spin()
        else:
            if isinstance(initial_state, dict):
                initial_state = [initial_state[k] for k in model.indices]
            _init_state = np.array(initial_state)

            if structure == None:
                # validate initial_state size
                if len(initial_state) != ising_graph.size():
                    raise ValueError(
                        "the size of the initial state should be {}"
                        .format(ising_graph.size()))
            else:
                # resize _initial_state
                temp_state = [1]*int(structure['size'])
                for k,ind in enumerate(model.indices):
                    temp_state[structure['dict'][ind]] = _init_state[k]
                _init_state = temp_state


            def _generate_init_state(): return np.array(_init_state)
        # -------------------------------- make init state generator

        # choose updater -------------------------------------------
        _updater_name = updater.lower().replace('_', '').replace(' ', '')
        if _updater_name not in self._make_system:
            raise ValueError('updater is one of "single spin flip or swendsen wang"')
        algorithm = self._algorithm[_updater_name]
        sa_system = self._make_system[_updater_name](_generate_init_state(), ising_graph)
        # ------------------------------------------- choose updater
        response = self._cxxjij_sampling(
            model, _generate_init_state,
            algorithm, sa_system,
            reinitialize_state, seed, structure
        )

        response.info['schedule'] = self.schedule_info

        return response

    def sample_hubo(self, interactions: list, var_type,
                    beta_min=None, beta_max=None, schedule=None,
                    num_sweeps=100, num_reads=1,
                    init_state=None, seed=None):
        """sampling from higher order unconstrainted binary optimization.

        Args:
            interactions (list): ordered by degree of interaction.
            var_type (str, openjij.VarType): "SPIN" or "BINARY"
            beta_min (float, optional): Minimum beta (initial inverse temperature). Defaults to None.
            beta_max (float, optional): Maximum beta (final inverse temperature). Defaults to None.
            schedule (list, optional): schedule list. Defaults to None.
            num_sweeps (int, optional): number of sweeps. Defaults to 100.
            num_reads (int, optional): number of reads. Defaults to 1.
            init_state (list, optional): initial state. Defaults to None.
            seed (int, optional): [description]. Defaults to None.

        Returns:
            :class:`openjij.sampler.response.Response`: results

        Examples::

            >>> sampler = oj.SASampler()
            >>> h = {0: -1}
            >>> J = {(0, 1): -1}
            >>> K = {(0, 1, 2): 1}
            >>> response = sampler.sample_hubo([h, J, K], var_type="SPIN")

        """

        self._setting_overwrite(
            beta_min=beta_min, beta_max=beta_max,
            num_sweeps=num_sweeps, num_reads=num_reads
        )

        bhom = BinaryHigherOrderModel(interactions)

        if schedule or self.schedule:
            self._schedule = self._convert_validation_schedule(
                schedule if schedule else self.schedule
            )
            self.schedule_info = {'schedule': 'custom schedule'}
        else:
            schedule = default_schedule(
                bhom,
                beta_min=self._schedule_setting['beta_min'],
                beta_max=self._schedule_setting['beta_max'],
                num_sweeps=self._schedule_setting['num_sweeps'])
            self.schedule_info = {
                'beta_max': schedule[-1][0],
                'beta_min': schedule[0][0],
                'num_sweeps': self._schedule_setting['num_sweeps']
            }

        response = hubo_sa_sampling(
            bhom, var_type,
            schedule=schedule, schedule_info=self.schedule_info,
            num_sweeps=self._schedule_setting['num_sweeps'],
            num_reads=self._schedule_setting['num_reads'],
            init_state=init_state, seed=seed
        )
        return response


def geometric_ising_beta_schedule(model: openjij.model.BinaryQuadraticModel,
                                  beta_max=None, beta_min=None,
                                  num_sweeps=1000):
    """make geometric cooling beta schedule

    Args:
        model (openjij.BinaryQuadraticModel)
        beta_max (float, optional): [description]. Defaults to None.
        beta_min (float, optional): [description]. Defaults to None.
        num_sweeps (int, optional): [description]. Defaults to 1000.
    Returns:
        list of cxxjij.utility.ClassicalSchedule, list of beta range [max, min]
    """
    if beta_min is None or beta_max is None:
        # generate Ising matrix
        ising_interaction = model.interaction_matrix()
        if (model.vartype == openjij.BINARY):
            # convert to ising matrix
            qubo_to_ising(ising_interaction)

        ising_interaction = np.abs(ising_interaction)

        #automatical setting of min, max delta energy
        abs_bias = np.sum(ising_interaction, axis=1)

        min_delta_energy = np.min(ising_interaction[ising_interaction > 0])
        max_delta_energy = np.max(abs_bias[abs_bias > 0])

    # TODO: More optimal schedule ?

    beta_min = np.log(2) / max_delta_energy if beta_min is None else beta_min
    beta_max = np.log(100) / min_delta_energy if beta_max is None else beta_max

    num_sweeps_per_beta = max(1, num_sweeps // 1000)

    # set schedule to cxxjij
    schedule = cxxjij.utility.make_classical_schedule_list(
        beta_min=beta_min, beta_max=beta_max,
        one_mc_step=num_sweeps_per_beta,
        num_call_updater=num_sweeps
    )

    return schedule, [beta_max, beta_min]



import unittest

from openjij.sampler.cmos_annealer import _cmos_annealer_format
from openjij import CMOSAnnealer

TOKEN = ''

if len(TOKEN) > 0:

    class TestCMOS(unittest.TestCase):
        def test_cmos_format(self):
            J = {(0, 1): -1, (1, 2): -0.5}
            h = {0: -1.5, 2: -1.2}
            X, Y = 5, 5
            max_abs = 3
            result = _cmos_annealer_format(h, J, X, Y, max_abs)
            self.assertListEqual(
                result, [[0, 0, 1, 0, -2], [1, 0, 2, 0, -1],
                         [0, 0, 0, 0, -3], [2, 0, 2, 0, -2]]
            )

            # FPGA case
            X, Y = 80, 80
            h, J = {X*Y-1: -1}, {(0, 1): -1, (X*Y-2, X*Y-1): -1}
            _ = _cmos_annealer_format(h, J, X, Y, 3)

        def test_cmos_sampler(self):
            sampler = CMOSAnnealer(token=TOKEN)
            h, J = {0: 1}, {(0, 1): -2, (1, 2): -1}
            num_reads = 10
            res = sampler.sample_ising(h, J, num_reads=num_reads)
            self.assertEqual(len(res.record), num_reads)

            sampler = CMOSAnnealer(token=TOKEN, num_reads=num_reads)
            res = sampler.sample_ising(h, J, num_reads=num_reads)
            self.assertEqual(len(res.record), num_reads)

            # QUBO Test
            q = {(0, 0): -1, (0, 1): -1, (1, 2): 1}
            res = sampler.sample_qubo(q)
            self.assertEqual(res.first.sample, {0: 1, 1: 1, 2: 0})

            res = sampler.sample_ising(h, J, beta_min=0.1, beta_max=10.0)
            self.assertEqual(res.info['schedule']['beta_min'], 0.1)
            self.assertEqual(res.info['schedule']['beta_max'], 10.0)

        def _grid_test(self, X, Y, sampler):
            h, J = {X*Y-1: -1}, {(0, 1): -1, (X*Y-2, X*Y-1): -1}
            _ = sampler.sample_ising(h, J)  # validate Ising
            h, J = disconnect_in_Kings()
            with self.assertRaises(ValueError):
                _ = sampler.sample_ising(h, J)
            h, J = over_grid(X, Y)
            with self.assertRaises(ValueError):
                _ = sampler.sample_ising(h, J)

        def test_gird_validate(self):
            sampler = CMOSAnnealer(token=TOKEN)
            # check ASIC Grid 352 * 176
            self._grid_test(X=352, Y=176, sampler=sampler)

            # check FPGA Grid 80*80
            sampler = CMOSAnnealer(token=TOKEN, machine_type='FPGA')
            self._grid_test(X=80, Y=80, sampler=sampler)

        def test_parameter_validate(self):
            h, J = {0: -1}, {(0, 1): -1, (1, 2): -1}
            sampler = CMOSAnnealer(token=TOKEN)
            with self.assertRaises(ValueError):
                _ = sampler.sample_ising(h, J, num_reads=15)
            with self.assertRaises(ValueError):
                _ = sampler.sample_ising(h, J, num_sweeps=101)
            with self.assertRaises(ValueError):
                _ = sampler.sample_ising(h, J, beta_max=0)
            with self.assertRaises(ValueError):
                _ = sampler.sample_ising(h, J, beta_max=0)
            with self.assertRaises(ValueError):
                _ = sampler.sample_ising(h, J, beta_min=0)
            with self.assertRaises(ValueError):
                _ = sampler.sample_ising(h, J, step_length=1001)

        # def test_embbed(self):
        #     from dwave.system.composites import EmbeddingComposite
        #     sampler = EmbeddingComposite(CMOSAnnealer(token=TOKEN))
        #     q = {('a', 'b'): -1, ('b', 'c'): 1, ('c', 'a'): 1}
        #     res = sampler.sample_qubo(q)
        #     self.assertEqual(res.first.sample, {'a': 1, 'b': 1, 'c': 0})

    def disconnect_in_Kings():
        J = {(0, 2): -1, (0, 1): 1}
        h = {0: 1}
        return h, J

    def over_grid(X, Y):
        J = {(0, 1): -1}
        h = {X*Y: -1}
        return h, J

    if __name__ == '__main__':
        unittest.main()
